
/*!
 *************************************************************************************
 * \file adaptive_loop_filter.c
 *
 * \brief
 *    adaptive loop filter routines
 *
 * \author
 *    Takeshi Chujoh         <takeshi.chujoh@toshiba.co.jp>             \n
 *    Akiyuki Tanizawa       <akiyuki.tanizawa@toshiba.co.jp>           \n
 *    Goki Yasuda            <goki1.yasuda@toshiba.co.jp>               \n
 *    Naofumi Wada           <naofumi.wada@toshiba.co.jp>               \n
 *    Takashi Watanabe       <takashi39.watanabe@toshiba.co.jp>         \n
 *************************************************************************************
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "adaptive_loop_filter.h"
#include "image.h"
#include "memalloc.h"
#include "vlc.h"
#include "cabac.h"
#include "biaridecod.h"
#include "context_ini.h"
#ifdef SWITCHED_FILTERS
#include "header.h"
#include "elements.h"
#include "fmo.h"
#endif

#ifdef ADAPTIVE_LOOP_FILTER

//------- global parameters for adaptive_loop_filter.c
int    num_bit;
int    r_offset;
int    ext_offset;
int    base_type;

imgpel **imgY_ext;
imgpel ***imgUV_ext;
int    **alf_coef_base;
int    ***alf_coef_tap_base;
int    **alf_coef_base_chroma;

int    alf_blk_size[8] = { 8, 16, 24, 32, 48, 64, 96, 128 };

int filter_base_luma[42] = {
  0,  0,  0,  0,   0,  0,  0,  0, 0,
  0,  0,  0, -1,   5, -1,  0,  0, 0,
  0,  0,  0,  0,  -7,  0,  0,  0, 0,
  0, -1,  0, -5,  18, -5,  0, -1, 0,
  0,  5, -7, 18, 219,  0
};
int filter_base_chroma[14] = {
   5, -5,  -5, -5,  5,
  -5,  0,  40,  0, -5,
  -5, 40, 160,  0
};

int filter_pos5x5[14] = {
  20, 21, 22, 23, 24,
  29, 30, 31, 32, 33,
  38, 39, 40, 41
};
int filter_pos7x7[26] = {
  10, 11, 12, 13, 14, 15, 16,
  19, 20, 21, 22, 23, 24, 25,
  28, 29, 30, 31, 32, 33, 34,
  37, 38, 39, 40, 41
};
int filter_pos9x9[42] = {
   0,  1,  2,  3,  4,  5,  6,  7,  8, 
   9, 10, 11, 12, 13, 14, 15, 16, 17,
  18, 19, 20, 21, 22, 23, 24, 25, 26,
  27, 28, 29, 30, 31, 32, 33, 34, 35,
  36, 37, 38, 39, 40, 41
};

int symmetric_array9x9[81] = {
   0,  1,  2,  3,  4,  5,  6,  7,  8,
   9, 10, 11, 12, 13, 14, 15, 16, 17,    
  18, 19, 20, 21, 22, 23, 24, 25, 26,
  27, 28, 29, 30, 31, 32, 33, 34, 35,       
  36, 37, 38, 39, 40, 39, 38, 37, 36, 
  35, 34, 33, 32, 31, 30, 29, 28, 27, 
  26, 25, 24, 23, 22, 21, 20, 19, 18,
  17, 16, 15, 14, 13, 12, 11, 10,  9, 
   8,  7,  6,  5,  4,  3,  2,  1,  0
};
int symmetric_array7x7[49] = {
   0,  1,  2,  3,  4,  5,  6,    
   7,  8,  9, 10, 11, 12, 13,
  14, 15, 16, 17, 18, 19, 20,     
  21, 22, 23, 24, 23, 22, 21, 
  20, 19, 18, 17, 16, 15, 14, 
  13, 12, 11, 10,  9,  8,  7,
   6,  5,  4,  3,  2,  1,  0, 
};
int symmetric_array5x5[25] = {    
   0,  1,  2,  3,  4, 
   5,  6,  7,  8,  9,        
  10, 11, 12, 11, 10,  
   9,  8,  7,  6,  5,  
   4,  3,  2,  1,  0,  
};
int filter_base_flag[3][4];
//------- end

/*! 
*************************************************************************************
* \brief
*    Allocates memory for a alfps
*
* \return
*    pointer to a alfps
*************************************************************************************
*/
alf_parameter_set_rbsp_t *AllocALFPS (int *memory_size)
{
  alf_parameter_set_rbsp_t *p;

  if ((p=calloc (sizeof (alf_parameter_set_rbsp_t), 1)) == NULL)
    no_mem_exit ("AllocALFPS: ALFPS");
  *memory_size += sizeof(alf_parameter_set_rbsp_t);

  *memory_size += get_mem1Dint(&(p->alf.coeff), MAX_NUM_COEF);
  *memory_size += get_mem1Dint(&(p->alf.coeff_chroma), MAX_NUM_COEF_C);

  img->max_WidthInAlfBlks  = img->width  / alf_blk_size[0];
  img->max_HeightInAlfBlks = img->height / alf_blk_size[0];
  if( (img->width  % alf_blk_size[0])!=0 ) img->max_WidthInAlfBlks++;
  if( (img->height % alf_blk_size[0])!=0 ) img->max_HeightInAlfBlks++;
  img->max_num_alf_block   = img->max_WidthInAlfBlks * img->max_HeightInAlfBlks;
  if(((p->alf_blk_data) = (AlfBlock *) calloc(img->max_num_alf_block, sizeof(AlfBlock))) == NULL)
    no_mem_exit("AllocALFPS: p->alf_blk_data");
  *memory_size += sizeof(AlfBlock)*img->max_num_alf_block;

  return p;
}

/*! 
*************************************************************************************
* \brief
*    Frees a alfps
*
* \param alfps
*     alfps to be freed
*
* \return
*    none
*************************************************************************************
*/
void FreeALFPS (alf_parameter_set_rbsp_t *p)
{
  assert(p != NULL);

  if(p->alf.coeff != NULL)
  {
    free_mem1Dint (p->alf.coeff);
    p->alf.coeff=NULL;
  }

  if(p->alf.coeff != NULL)
  {
    free_mem1Dint (p->alf.coeff_chroma);
    p->alf.coeff_chroma=NULL;
  }

  if(p->alf_blk_data != NULL)
  {
    free(p->alf_blk_data);
    p->alf_blk_data=NULL;
  }

  free (p);
  p=NULL;
}

/*!
************************************************************************
* \brief
*    Dynamic memory allocation of frame size related global buffers
*    buffers are defined in global.h, allocated memory must be freed in
*    void FreeALFGlobalBurrers()
*
*  \par Output:
*     Number of allocated bytes
***********************************************************************
*/
int InitALFGlobalBuffers(void)
{
  int memory_size = 0;

  memory_size += get_mem2Dpel (&(imgY_ext), img->height+MAX_NUM_TAP, img->width+MAX_NUM_TAP);
  memory_size += get_mem3Dpel (&(imgUV_ext), 2, img->height_cr+MAX_NUM_TAP_C, img->width_cr+MAX_NUM_TAP_C);
  memory_size += get_mem2Dint (&(alf_coef_base), 4, MAX_NUM_COEF);
  memory_size += get_mem3Dint (&(alf_coef_tap_base), 3, 4, MAX_NUM_COEF);
  memory_size += get_mem2Dint (&(alf_coef_base_chroma), 4, MAX_NUM_COEF_C);

  // allocate ALF parameter set
  active_alfps = AllocALFPS(&memory_size);

  return memory_size;
}

/*!
************************************************************************
* \brief
*    Free allocated memory of frame size related global buffers
*    buffers are defined in global.h, allocated memory is allocated in
*    int InitALFGlobalBuffers()
************************************************************************
*/
void FreeALFGlobalBurrers(void)
{
  free_mem2Dpel (imgY_ext);
  free_mem3Dpel (imgUV_ext, 2);
  free_mem2Dint (alf_coef_base);
  free_mem3Dint (alf_coef_tap_base, 3);
  free_mem2Dint (alf_coef_base_chroma);

  FreeALFPS (active_alfps);
}

/*!
 *****************************************************************************************
 * \brief
 *    set ALF parameters (alfps -> p)
 *****************************************************************************************
 */
void SetALFParameters(StorablePicture *p, alf_parameter_set_rbsp_t *alfps)
{
  p->alf_flag = alfps->alf_flag;

  if (alfps->alf_flag)
  {
    p->alf.tap           = alfps->alf.tap;
    p->alf.num_coeff     = alfps->alf.num_coeff;
    memcpy(p->alf.coeff, alfps->alf.coeff, sizeof(int)*(p->alf.num_coeff));

    p->alf_chroma_idc = alfps->alf_chroma_idc;
    if (p->alf_chroma_idc)
    {
      p->alf.tap_chroma       = alfps->alf.tap_chroma;
      p->alf.num_coeff_chroma = alfps->alf.num_coeff_chroma;
      memcpy(p->alf.coeff_chroma, alfps->alf.coeff_chroma, sizeof(int)*(p->alf.num_coeff_chroma));
    }

    p->block_control_flag = alfps->block_control_flag;
    if (p->block_control_flag)
    {
      p->alf_block_size_idx = alfps->alf_block_size_idx;
      p->alf_block_size     = alfps->alf_block_size;
      p->WidthInAlfBlks     = alfps->WidthInAlfBlks;
      p->HeightInAlfBlks    = alfps->HeightInAlfBlks;
      p->num_alf_block      = alfps->num_alf_block;
      memcpy( p->alf_blk_data, alfps->alf_blk_data, sizeof(AlfBlock)*(p->num_alf_block));
    }
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    make temporally frame for filtering
 *****************************************************************************************
 */
void MakeExtendFrame(imgpel** ImgDec, imgpel** ImgExt, int width, int height, int offset)
{
  int x, y;

  for(y=0; y<offset; y++)
  {
    for(x=0; x<offset; x++)
    {
      ImgExt[y][x] = ImgDec[0][0];
    }
    for(x=offset; x<width+offset; x++)
    {
      ImgExt[y][x] = ImgDec[0][x-offset];
    }
    for(x=width+offset; x<width+(offset<<1); x++)
    {
      ImgExt[y][x] = ImgDec[0][width-1];
    }
  }
  for(y=offset; y<height+offset; y++)
  {
    for(x=0; x<offset; x++)
    {
      ImgExt[y][x] = ImgDec[y-offset][0];
    }
    for(x=offset; x<width+offset; x++)
    {
      ImgExt[y][x] = ImgDec[y-offset][x-offset];
    }
    for(x=width+offset; x<width+(offset<<1); x++)
    {
      ImgExt[y][x] = ImgDec[y-offset][width-1];
    }
  }
  for(y=height+offset; y<height+(offset<<1); y++)
  {
    for(x=0; x<offset; x++)
    {
      ImgExt[y][x] = ImgDec[height-1][0];
    }
    for(x=offset; x<width+offset; x++)
    {
      ImgExt[y][x] = ImgDec[height-1][x-offset];
    }
    for(x=width+offset; x<width+(offset<<1); x++)
    {
      ImgExt[y][x] = ImgDec[height-1][width-1];
    }
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    set restored frame
 *****************************************************************************************
 */
void SetRestFrame(imgpel** ImgRest, imgpel** ImgDec, int width, int height)
{
  int i;

  for(i=0; i<height; i++)
    memcpy(ImgDec[i], ImgRest[i], sizeof(imgpel)*width);
}

/*!
 *****************************************************************************************
 * \brief
 *    get base type for filter subtraction 
 *****************************************************************************************
 */
int GetALFBaseType(int slice_type, int ref)
{
  int type=0;

  if ( (slice_type==I_SLICE) || (slice_type==SI_SLICE) )
  {
    type = 0;
  }
  else if ( (slice_type==P_SLICE) || (slice_type==SP_SLICE) )
  {
    type = 1;
  }
  else
  {
    type = (ref>0) ? 2 : 3;
  }

  return type;
}

/*!
 *****************************************************************************************
 * \brief
 *    initialize base of filter coefficients
 *****************************************************************************************
 */
void InitFilterBase()
{
  int i, tap_idx, type;

  // luma filter base
  for(tap_idx=0; tap_idx<3; tap_idx++)
  {
    for(type=0; type<4; type++)
    {
      for(i=0; i<MAX_NUM_COEF; i++)
      {
        alf_coef_base[type][i] = filter_base_luma[i];
      }
      filter_base_flag[tap_idx][type] = 0;
    }
  }

  // chroma filter base
  for(type=0; type<4; type++)
  {
    for(i=0; i<MAX_NUM_COEF_C; i++)
    {
      alf_coef_base_chroma[type][i] = filter_base_chroma[i];
    }
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    set of adaptive loop filter coefficients 
 *****************************************************************************************
 */
void SetALFCoef(StorablePicture *p)
{
  int i, N, tap;
  int *filter_pos;
  int *base_flag;
  int **base_coef;

  tap = p->alf.tap;
  N   = (tap*tap+1)>>1;

  switch(p->alf.tap)
  {
  case 5:
    filter_pos = filter_pos5x5;
    base_flag  = filter_base_flag[0];
    base_coef  = alf_coef_tap_base[0];
    break;
  case 7:
    filter_pos = filter_pos7x7;
    base_flag  = filter_base_flag[1];
    base_coef  = alf_coef_tap_base[1];
    break;
  case 9:
    filter_pos = filter_pos9x9;
    base_flag  = filter_base_flag[2];
    base_coef  = alf_coef_tap_base[2];
    break;
  default:
    printf("not support filter size!!\n");
    exit(1);
    break;
  }

  if(base_flag[base_type] == 0)
  {
    for(i=0; i<N; i++)
    {
      p->alf.coeff[i] += alf_coef_base[base_type][filter_pos[i]];
      alf_coef_base[base_type][filter_pos[i]] = p->alf.coeff[i];
      base_coef[base_type][i]                 = p->alf.coeff[i];
    }
    base_flag[base_type] = 1;
  }
  else
  {
    for(i=0; i<N; i++)
    {
      p->alf.coeff[i] += base_coef[base_type][i];
      base_coef[base_type][i] = p->alf.coeff[i];
    }
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    set of adaptive loop-filter coefficients for chroma
 *****************************************************************************************
 */
void SetALFCoefChroma(StorablePicture *p)
{
  int i, tap, N;

  tap = p->alf.tap_chroma;
  N   = (tap*tap+1)>>1;

  for(i=0; i<N; i++)
  {
    p->alf.coeff_chroma[i] += alf_coef_base_chroma[base_type][i];
    alf_coef_base_chroma[base_type][i] = p->alf.coeff_chroma[i];
  }
}


//------- filtering for luma
/*!
 *****************************************************************************************
 * \brief
 *    adaptive loop filter main process
 *****************************************************************************************
 */
void AdaptiveLoopFilterProcess(StorablePicture *p)
{
  imgpel** ImgDec;
  imgpel** ImgRest;

  // initialize base coefficients
  if(p->idr_flag)
  {
    InitFilterBase();
  }

  if(p->alf_flag)
  {
    // set global variables
    num_bit    = NUM_BIT_SHIFT;
    r_offset   = 1<<(num_bit-1);
    ext_offset = p->alf.tap>>1;
    base_type  = GetALFBaseType(p->slice_type, p->used_for_reference);

    // set image parameters for Block-based ALF
    img->alf_block_size  = p->alf_block_size;
    img->num_alf_block   = p->num_alf_block;
    img->WidthInAlfBlks  = p->WidthInAlfBlks;
    img->HeightInAlfBlks = p->HeightInAlfBlks;
    img->alf_blk_data    = p->alf_blk_data;

    // extend image for filtering
    MakeExtendFrame( p->imgY, imgY_ext, img->width, img->height, ext_offset );

    // set image buffer pointer
    ImgDec  = imgY_ext;
    ImgRest = p->imgY;

    // set filter coefficients
    SetALFCoef(p);

    // Adaptive in-loop wiener filtering
    FilteringFrame( p, ImgDec, ImgRest  );
    
    // Adaptive in-loop wiener filtering for chroma
    if(p->alf_chroma_idc)
    {
      AdaptiveLoopFilterProcessChroma(p);
    }
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    filtering frame for luma
 *****************************************************************************************
 */
void FilteringFrame(StorablePicture *p, imgpel** ImgDec, imgpel** ImgRest)
{
  if(p->block_control_flag)
  {
    BlockAdaptiveFilteringProcess( p, ImgDec, ImgRest );
  }
  else
  {
    FilteringProcess( ImgDec, ImgRest, p->alf.coeff, p->alf.tap );
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    filtering process
 *****************************************************************************************
 */
void FilteringProcess(imgpel** ImgDec, imgpel** ImgRest, int *qh, int tap)
{
  int i, x, y, value, N, offset;
  int *pFiltPos, PixSum[MAX_NUM_COEF];
  int yy, xx;

  N      = (tap*tap+1)>>1;
  offset = tap>>1;

  switch(tap)
  {
  case 9:
    pFiltPos = symmetric_array9x9;
    break;
  case 7:
    pFiltPos = symmetric_array7x7;
    break;
  case 5:
    pFiltPos = symmetric_array5x5;
    break;
  default:
    printf("not support filter size : %d", tap);
    exit(1);
    break;
  }

  for(y=ext_offset; y<img->height+ext_offset; y++)
  {
    for(x=ext_offset; x<img->width+ext_offset; x++)
    {
      i=0;
      memset(PixSum, 0, sizeof(int)*N);
      for(yy=y-offset; yy<=y+offset; yy++)
      {
        for(xx=x-offset; xx<=x+offset; xx++)
        {
          PixSum[pFiltPos[i]] += (int)ImgDec[yy][xx];
          i++;
        }
      }
      value = 0;
      for(i=0; i<N; i++)
      {
        value += qh[i]*PixSum[i];
      }
      // DC offset
      value += qh[N]<<(img->bitdepth_luma-8);

      value = (value + r_offset)>>num_bit;
      ImgRest[y-ext_offset][x-ext_offset] = (imgpel)Clip1(value);
    }
  }
}


//------- filtering for chroma
/*!
 *****************************************************************************************
 * \brief
 *    adaptive loop filter main process for chroma
 *****************************************************************************************
 */
void AdaptiveLoopFilterProcessChroma(StorablePicture *p)
{
  imgpel ***ImgUVDec;
  imgpel ***ImgUVRest;

  // set global variables for chroma
  ext_offset = p->alf.tap_chroma>>1;
  base_type  = GetALFBaseType(p->slice_type, p->used_for_reference);

  // extend image for filtering
  MakeExtendFrame( p->imgUV[0], imgUV_ext[0], img->width_cr, img->height_cr, ext_offset );
  MakeExtendFrame( p->imgUV[1], imgUV_ext[1], img->width_cr, img->height_cr, ext_offset );

  // set image buffer pointer
  ImgUVDec  = imgUV_ext;
  ImgUVRest = p->imgUV;

  // set filter coefficients for chroma
  SetALFCoefChroma(p);

  // Adaptive in-loop wiener filtering for chroma
  FilteringFrameChroma( p, ImgUVDec, ImgUVRest );
}

/*!
 *****************************************************************************************
 * \brief
 *    filtering frame for chroma
 *****************************************************************************************
 */
void FilteringFrameChroma(StorablePicture *p, imgpel ***ImgUVDec, imgpel ***ImgUVRest)
{
  if((p->alf_chroma_idc>>1)&0x01)
  {
    FilteringProcessChroma(ImgUVDec, ImgUVRest, p->alf.coeff_chroma, p->alf.tap_chroma, 0);
  }
  if(p->alf_chroma_idc&0x01)
  {
    FilteringProcessChroma(ImgUVDec, ImgUVRest, p->alf.coeff_chroma, p->alf.tap_chroma, 1);
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    filtering process for chroma
 *****************************************************************************************
 */
void FilteringProcessChroma(imgpel ***ImgUVDec, imgpel ***ImgUVRest, int *qh, int tap, int chroma_id)
{
  int i, x, y, value, N, offset;
  int *pFiltPos, PixSum[MAX_NUM_COEF_C];
  int yy, xx;

  N      = (tap*tap+1)>>1;
  offset = tap>>1;

  switch(tap)
  {
  case 9:
    pFiltPos = symmetric_array9x9;
    break;
  case 7:
    pFiltPos = symmetric_array7x7;
    break;
  case 5:
    pFiltPos = symmetric_array5x5;
    break;
  default:
    printf("not support filter size : %d", tap);
    exit(1);
    break;
  }

  for(y=ext_offset; y<img->height_cr+ext_offset; y++)
  {
    for(x=ext_offset; x<img->width_cr+ext_offset; x++)
    {
      i=0;
      memset(PixSum, 0, sizeof(int)*N);
      for(yy=y-offset; yy<=y+offset; yy++)
      {
        for(xx=x-offset; xx<=x+offset; xx++)
        {
          PixSum[pFiltPos[i]] += (int)ImgUVDec[chroma_id][yy][xx];
          i++;
        }
      }
      value = 0;
      for(i=0; i<N; i++)
      {
        value += qh[i]*PixSum[i];
      }
      // DC offset
      value += qh[N]<<(img->bitdepth_chroma-8);

      value = (value + r_offset)>>num_bit;
      ImgUVRest[chroma_id][y-ext_offset][x-ext_offset] = (imgpel)Clip1_Chr(value);
    }
  }
}


//------- filtering based block
/*!
 *****************************************************************************************
 * \brief
 *    block-based adaptive loop filtering
 *****************************************************************************************
 */
void BlockAdaptiveFilteringProcess(StorablePicture *p, imgpel** ImgDec, imgpel** ImgRest)
{
  int i;
  AlfBlock *currAlfBlk;

  for (i=0; i<p->num_alf_block; i++)
  { 
    currAlfBlk = &p->alf_blk_data[i];

    if(currAlfBlk->alf_blk_flag)
    {
      FilteringProcessBlock(i, ImgDec, ImgRest, p->alf.coeff, p->alf.tap, p->alf_block_size, p->WidthInAlfBlks);
    }
    else
    {
      CopyDecToRestBlock(i, p->imgY, ImgRest, p->alf_block_size, p->WidthInAlfBlks);
    }
  }
}
/*!
 *****************************************************************************************
 * \brief
 *    filtering process per block
 *****************************************************************************************
 */
void FilteringProcessBlock(int blk_addr, imgpel** ImgDec, imgpel** ImgRest, int *qh, int tap, int blk_size, int width_in_alf_blks)
{
  int i, value, offset, N;
  int x, y, blk_x, blk_y, max_blk_x, max_blk_y;
  int *pFiltPos, PixSum[MAX_NUM_COEF];
  int yy, xx;

  N      = (tap*tap+1)>>1;
  offset = tap>>1;

  switch(tap)
  {
  case 9:
    pFiltPos = symmetric_array9x9;
    break;
  case 7:
    pFiltPos = symmetric_array7x7;
    break;
  case 5:
    pFiltPos = symmetric_array5x5;
    break;
  default:
    printf("not support filter size : %d", tap);
    exit(1);
    break;
  }

  GetSamplePosForBALF(blk_addr, &blk_x, &blk_y, blk_size, width_in_alf_blks);

  max_blk_x = blk_x + blk_size;
  max_blk_y = blk_y + blk_size;
  if(max_blk_x > img->width)  max_blk_x = img->width;
  if(max_blk_y > img->height) max_blk_y = img->height;

  for(y=blk_y+ext_offset; y<max_blk_y+ext_offset; y++)
  {
    for(x=blk_x+ext_offset; x<max_blk_x+ext_offset; x++)
    {
      i=0;
      memset(PixSum, 0, sizeof(int)*N);
      for(yy=y-offset; yy<=y+offset; yy++)
      {
        for(xx=x-offset; xx<=x+offset; xx++)
        {
          PixSum[pFiltPos[i]] += (int)ImgDec[yy][xx];
          i++;
        }
      }
      value = 0;
      for(i=0; i<N; i++)
      {
        value += qh[i]*PixSum[i];
      }
      // DC offset
      value += qh[N]<<(img->bitdepth_luma-8);

      value = (value + r_offset)>>num_bit;
      ImgRest[y-ext_offset][x-ext_offset] = (imgpel)Clip1(value);
    }
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    copy from dec block to rest block
 *****************************************************************************************
 */
void CopyDecToRestBlock(int blk_addr, imgpel** ImgDec, imgpel** ImgRest, int blk_size, int width_in_alf_blks)
{
  int x, y, blk_x, blk_y, max_blk_x, max_blk_y;

  GetSamplePosForBALF(blk_addr, &blk_x, &blk_y, blk_size, width_in_alf_blks);

  max_blk_x = blk_x + blk_size;
  max_blk_y = blk_y + blk_size;
  if(max_blk_x > img->width)  max_blk_x = img->width;
  if(max_blk_y > img->height) max_blk_y = img->height;

  for(y=blk_y; y<max_blk_y; y++)
  {
    for(x=blk_x; x<max_blk_x; x++)
    {
      ImgRest[y][x] = ImgDec[y][x]; 
    }
  }
}

/*!
 ************************************************************************
 * \brief
 *    returns the x and y block coordinates for a given AlfBlockAddress
 ************************************************************************
 */
void GetBlockPosForBALF(int blk_addr, int *x, int*y, int width_in_blks)
{
  *x = (blk_addr % width_in_blks);
  *y = (blk_addr / width_in_blks);
}

/*!
 ************************************************************************
 * \brief
 *    returns the x and y sample coordinates for a given AlfBlockAddress
 ************************************************************************
 */
void GetSamplePosForBALF(int blk_addr, int *x, int*y, int blk_size, int width_in_blks)
{
  GetBlockPosForBALF(blk_addr, x, y, width_in_blks);
  
  (*x) *= blk_size;
  (*y) *= blk_size;
}

/*!
 ************************************************************************
 * \brief
 *    returns 1 if the ALF block at the given address is available
 ************************************************************************
 */
int ALFBlockIsAvailable(int blk_addr)
{
  if ((blk_addr < 0) || (blk_addr > ((int)img->num_alf_block - 1)))
    return 0;
  
  return 1;
}

/*!
 ************************************************************************
 * \brief
 *    Checks the availability of neighboring ALF blocks of
 *    the current ALF block for prediction and context determination;
 ************************************************************************
 */
void CheckAvailabilityOfNeighborsAlfBlock()
{
  const int alfb_nr = img->current_alfb_nr;
  AlfBlock *currAlfBlk = &img->alf_blk_data[alfb_nr];

  // mark all neighbors as unavailable
  currAlfBlk->alfb_available_up   = NULL;
  currAlfBlk->alfb_available_left = NULL;

  currAlfBlk->alfbAddrA = alfb_nr - 1;
  currAlfBlk->alfbAddrB = alfb_nr - img->WidthInAlfBlks;
  currAlfBlk->alfbAddrC = alfb_nr - img->WidthInAlfBlks + 1;
  currAlfBlk->alfbAddrD = alfb_nr - img->WidthInAlfBlks - 1;

  currAlfBlk->alfbAvailA = ALFBlockIsAvailable(currAlfBlk->alfbAddrA) && (( alfb_nr    % img->WidthInAlfBlks)!=0);
  currAlfBlk->alfbAvailB = ALFBlockIsAvailable(currAlfBlk->alfbAddrB);
  currAlfBlk->alfbAvailC = ALFBlockIsAvailable(currAlfBlk->alfbAddrC) && (((alfb_nr+1) % img->WidthInAlfBlks)!=0);
  currAlfBlk->alfbAvailD = ALFBlockIsAvailable(currAlfBlk->alfbAddrD) && (( alfb_nr    % img->WidthInAlfBlks)!=0);
}

/*!
 ************************************************************************
 * \brief
 *    get neighbouring positions for non-aff coding
 * \param curr_alfb_nr
 *   current ALF block number (decoding order)
 * \param xN
 *    input x position
 * \param yN
 *    input y position
 * \param pix
 *    returns position informations
 ************************************************************************
 */
void GetNonAffNeighbourAlfBlock(unsigned int curr_alfb_nr, int xN, int yN, PixelPos *pix)
{
  AlfBlock *currAlfBlk = &img->alf_blk_data[curr_alfb_nr];
  int maxW, maxH;

  maxW = img->alf_block_size;
  maxH = img->alf_block_size;

  if ((xN<0)&&(yN<0))
  {
    pix->mb_addr   = currAlfBlk->alfbAddrD;
    pix->available = currAlfBlk->alfbAvailD;
  }
  else if ((xN<0)&&((yN>=0)&&(yN<maxH)))
  {
    pix->mb_addr   = currAlfBlk->alfbAddrA;
    pix->available = currAlfBlk->alfbAvailA;
  }
  else if (((xN>=0)&&(xN<maxW))&&(yN<0))
  {
    pix->mb_addr   = currAlfBlk->alfbAddrB;
    pix->available = currAlfBlk->alfbAvailB;
  }
  else if (((xN>=0)&&(xN<maxW))&&((yN>=0)&&(yN<maxH)))
  {
    pix->mb_addr   = curr_alfb_nr;
    pix->available = 1;
  }
  else if ((xN>=maxW)&&(yN<0))
  {
    pix->mb_addr   = currAlfBlk->alfbAddrC;
    pix->available = currAlfBlk->alfbAvailC;
  }
  else 
  {
    pix->available = 0;
  }

  if (pix->available)
  {
    pix->x = (xN + maxW) % maxW;
    pix->y = (yN + maxH) % maxH;

    GetSamplePosForBALF(pix->mb_addr, &(pix->pos_x), &(pix->pos_y), img->alf_block_size, img->WidthInAlfBlks);

    pix->pos_x += pix->x;
    pix->pos_y += pix->y;
  }
}

/*!
 ************************************************************************
 * \brief
 *    get neighbouring positions. MB AFF is automatically used from img structure
 * \param curr_alfb_nr
 *   current ALF block number (decoding order)
 * \param xN
 *    input x position
 * \param yN
 *    input y position
 * \param pix
 *    returns position informations
 ************************************************************************
 */
void GetNeighbourAlfBlock(int curr_alfb_nr, int xN, int yN, PixelPos *pix)
{
  if (curr_alfb_nr<0)
    error ("GetNeighbourAlfBlock: invalid block number", 100);

  GetNonAffNeighbourAlfBlock(curr_alfb_nr, xN, yN, pix);
}


//------- read ALF parameters
/*!
 *****************************************************************************************
 * \brief
 *    read ALF parameters
 *****************************************************************************************
 */
int read_ALFParameterSet(alf_parameter_set_rbsp_t *alfps, DataPartition *p)
{
  Bitstream *s = p->bitstream;
  
  assert (p != NULL);
  assert (p->bitstream != NULL);
  assert (p->bitstream->streamBuffer != 0);

  // ALF parameters

  // filter parameter for luma
  read_ALFLumaParam(alfps, s);

  // filter parameter for chroma
  read_ALFChromaParam(alfps, s);

  // block control parameter for luma
  read_ALFBlockControlParam(alfps, p);

  // padding
  if (s->frame_bitoffset%8 != 0 )
  {
    while(s->frame_bitoffset%8 != 0)
    {
      u_1("ALFPS: padding bits", s);
    }
  }

  return 1;
}

/*!
 *****************************************************************************************
 * \brief
 *    read ALF parameters for luma
 *****************************************************************************************
 */
int read_ALFLumaParam(alf_parameter_set_rbsp_t *alfps, Bitstream *s)
{
  int pos, N;
  char string[100];

  // ALF parameters for luma
  alfps->alf.tap = ue_v("ALFPS: alf_tap_size", s);
  alfps->alf.tap = (alfps->alf.tap<<1) + 5;

  N = (alfps->alf.tap*alfps->alf.tap+1)>>1;
  N = N + 1;  // DC offset
  alfps->alf.num_coeff = N;

  // filter coefficients for luma
  for(pos=0; pos<N; pos++)
  {
    sprintf(string, "ALFPS: %d pos", pos);
    alfps->alf.coeff[pos] = se_v(string, s);
  }

  return 1;
}

/*!
 *****************************************************************************************
 * \brief
 *    read ALF parameters for chroma
 *****************************************************************************************
 */
int read_ALFChromaParam(alf_parameter_set_rbsp_t *alfps, Bitstream *s)
{
  int pos, N;
  char string[100];

  // ALF parameters for chroma
  alfps->alf_chroma_idc = u_v(2, "ALFPS: alf_chroma_idc", s);

  if(alfps->alf_chroma_idc)
  {
    alfps->alf.tap_chroma = ue_v("ALFPS: filter_tap_chroma", s);
    alfps->alf.tap_chroma = (alfps->alf.tap_chroma<<1) + 5;

    N = (alfps->alf.tap_chroma*alfps->alf.tap_chroma+1)>>1;
    N = N + 1;  // DC offset
    alfps->alf.num_coeff_chroma = N;

    // filter coefficients for chroma
    for(pos=0; pos<N; pos++)
    {
      sprintf(string, "ALFPS: [chroma] %d pos", pos);
      alfps->alf.coeff_chroma[pos] = se_v(string, s);
    }
  }

  return 1;
}

/*!
 *****************************************************************************************
 * \brief
 *    read ALF parameters for block-based
 *****************************************************************************************
 */
int read_ALFBlockControlParam(alf_parameter_set_rbsp_t *alfps, DataPartition *p)
{
  Bitstream *s = p->bitstream;

  // block-based on-off parameter
  alfps->block_control_flag = u_1("ALFPS: block_control_flag", s);

  if(alfps->block_control_flag)
  {
    alfps->alf_block_size_idx = u_v(3, "ALFPS: alf_block_size_idx", s);
    alfps->alf_block_size     = alf_blk_size[alfps->alf_block_size_idx];
    alfps->WidthInAlfBlks     = img->width  / alfps->alf_block_size;
    alfps->HeightInAlfBlks    = img->height / alfps->alf_block_size;
    if( (img->width  % alfps->alf_block_size)!=0 ) alfps->WidthInAlfBlks++;
    if( (img->height % alfps->alf_block_size)!=0 ) alfps->HeightInAlfBlks++;
    alfps->num_alf_block      = alfps->WidthInAlfBlks * alfps->HeightInAlfBlks;

    if (active_pps->entropy_coding_mode_flag == CABAC)
    {
      read_BlockFlagCABAC(alfps, p);

      if(alfps->block_control_flag)
      {
        s->frame_bitoffset = s->read_len<<3;
      }      
    }
    else
    {
      read_BlockFlagRunLength(alfps, s);
    }
  }

  return 1;
}

/*!
 *****************************************************************************************
 * \brief
 *    read block flag by Run-length coding
 *****************************************************************************************
 */
int read_BlockFlagRunLength(alf_parameter_set_rbsp_t *alfps, Bitstream *s)
{
  int i;
  int alf_blk_flag_run=0;
  Boolean prev_alf_blk_flag=FALSE;
  AlfBlock *currAlfBlk;

  alf_blk_flag_run = 0;

  for (i=0; i<alfps->num_alf_block; i++)
  {
    currAlfBlk = &alfps->alf_blk_data[i];

    if( i==0 )
    {
      currAlfBlk->alf_blk_flag = (Boolean)u_1("ALFPS: alf_blk_flag", s);
      alf_blk_flag_run         = ue_v("ALFPS: alf_blk_flag_run", s);
      prev_alf_blk_flag        = currAlfBlk->alf_blk_flag;
    }
    else
    {
      if(alf_blk_flag_run==0)
      {
        currAlfBlk->alf_blk_flag = (prev_alf_blk_flag) ? FALSE : TRUE;
        if(i!=alfps->num_alf_block-1)
        {
          alf_blk_flag_run = ue_v("ALFPS: alf_blk_flag_run", s);
          prev_alf_blk_flag = currAlfBlk->alf_blk_flag;
        }
      }
      else
      {
        currAlfBlk->alf_blk_flag = prev_alf_blk_flag;
        alf_blk_flag_run--;
      }
    }
  }

  return 1;
}

/*!
 *****************************************************************************************
 * \brief
 *    read block flag by CABAC
 *****************************************************************************************
 */
int read_BlockFlagCABAC(alf_parameter_set_rbsp_t *alfps, DataPartition *p)
{
  int i;
  int len = 0;
  AlfBlock *currAlfBlk;
  SyntaxElement currSE;
  int ByteStartPosition;
#if TRACE
  char string[100];
#endif
  Bitstream *s = p->bitstream;

  // set image paramters
  img->alf_block_size      = alfps->alf_block_size;
  img->WidthInAlfBlks      = alfps->WidthInAlfBlks;
  img->HeightInAlfBlks     = alfps->HeightInAlfBlks;
  img->num_alf_block       = alfps->num_alf_block;
  img->alf_blk_data        = alfps->alf_blk_data;

  ByteStartPosition = s->frame_bitoffset/8;
  if (s->frame_bitoffset%8 != 0)
  {
    ByteStartPosition++;
  }
  arideco_start_decoding_AlfBlkFlag (&p->de_cabac, s->streamBuffer, ByteStartPosition, &s->read_len);
  init_contexts (img);
  p->readSyntaxElement = readSyntaxElement_CABAC;

  for (i=0; i<alfps->num_alf_block; i++)
  {
    currAlfBlk = &alfps->alf_blk_data[i];

    img->current_alfb_nr = i;
    currAlfBlk = &alfps->alf_blk_data[i];
    currSE.type = SE_HEADER;

    CheckAvailabilityOfNeighborsAlfBlock();
    CheckAvailabilityOfNeighborsAlfBlockCABAC();
    currSE.reading = readAlfBlockFlagCABAC;
#if TRACE
    sprintf(string, "ALFPS: alf_blk_flag[%4d]", i);
    strncpy(currSE.tracestring,string,TRACESTRING_SIZE);
#endif
    p->readSyntaxElement(&currSE, img, input, p);
    len+=currSE.len;
    currAlfBlk->alf_blk_flag = (Boolean)currSE.value1;
  }

  return len;
}

#ifdef SWITCHED_FILTERS
/*!
 *****************************************************************************************
 * \brief
 *    read ALF parameters in SliceHeader
 *****************************************************************************************
 */
void read_ALFHeader()
{
  Slice *currSlice = img->currentSlice;
  int dP_nr = assignSE2partition[currSlice->dp_mode][SE_HEADER];
  DataPartition *partition = &(currSlice->partArr[dP_nr]);
  Bitstream *currStream = partition->bitstream;

  active_alfps->alf_flag = u_1("SH: adaptive_loop_filter_flag", currStream);

  // padding end of slice header for ALF
  if ( currStream->frame_bitoffset%8 != 0 )
  {
    while(currStream->frame_bitoffset%8 != 0)
    {
      u_1("SH: padding bits for ALF", currStream);
    }
  }

  // adaptive loopfilter parameters
  if(active_alfps->alf_flag)
  {
    read_ALFParameterSet(active_alfps, partition);
  }
}
#endif


#endif // ADAPTIVE_LOOP_FILTER
